@echo off &setlocal
rem Name of the .txt file (without file extension) which contains the pixel information
rem The expected format is: lines of HEX digits with 4-bit RGBI information
set "filename=redblock"

setlocal EnableDelayedExpansion
rem Create the pixel array first because we have to determine the dimensions anyway before the headers can be written
set "dgts=0123456789ABCDEF"
set /a "width=0,height=0,firstrow=1,d=170"
>"!filename!.hex" type nul
for /f "usebackq delims=" %%Y in ("!filename!.txt") do (
  set /a "height+=1"
  >"!filename!.hex.~tmp" (
    if defined firstrow (
      set "firstrow="
      for /f %%X in ('cmd /u /c ^<nul set /p "=%%Y"^|find /v ""') do (
        set /a "width+=1,value=0,c=0x%%X,i=(^!^!(c&8))*85,rgb=((^!^!(c&4)*d+i)<<16)|((^!^!(c&2)*d+i)<<8)|((c&1)*d+i)"
        for /l %%i in (1,1,6) do set /a "x=rgb&15,rgb>>=4" &for %%j in (!x!) do set "value=!dgts:~%%j,1!!value!"
        echo !value:~4,2! !value:~2,2! !value:~0,2!
      )
      rem The pixel rows may require a padding of 1..3 bytes because the number of bytes must be a multiple of 4 (DWORD alignment), yields 0 if no padding necessary
      set /a "pad=3-(width*3-1)%%4"
    ) else (
      set "line=_%%Y"
      for /l %%X in (1,1,!width!) do (
        set /a "value=0,c=0x!line:~%%X,1!,i=(^!^!(c&8))*85,rgb=((^!^!(c&4)*d+i)<<16)|((^!^!(c&2)*d+i)<<8)|((c&1)*d+i)"
        for /l %%i in (1,1,6) do set /a "x=rgb&15,rgb>>=4" &for %%j in (!x!) do set "value=!dgts:~%%j,1!!value!"
        echo !value:~4,2! !value:~2,2! !value:~0,2!
      )
    )
    rem Pad the pixel row (if necessary) to get it DWORD-aligned, the loop won't run if pad is 0
    for /l %%i in (1 1 !pad!) do echo 00
    echo(
  )
  rem The order of pixel rows must be bottom-up, prepend the new row to the existing data
  >nul copy /b "!filename!.hex.~tmp" + "!filename!.hex"
  >nul move /y "!filename!.hex.~tmp" "!filename!.hex"
)

rem 54 is the total size of the BITMAPFILEHEADER and BITMAPINFOHEADER at the beginning of the file, 3 is the number of bytes per pixel
set /a "filesize=54+(width*3+pad)*height"

rem Since we're about to write HEX-encoded data, convert integers to HEX strings
cmd /c exit !filesize!
set "filesize=!=exitcode!"
cmd /c exit !width!
set "width=!=exitcode!"
cmd /c exit !height!
set "height=!=exitcode!"

rem Finally we create the two header structures, and prepend them to the pixel array
rem Note that numbers consisting of multiple bytes have to be written in little endian byte order
>"!filename!.hex.~tmp" (
  rem BITMAPFILEHEADER https://docs.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-bitmapfileheader
  echo 42 4D &rem ASCII values of 'B' 'M'
  echo !filesize:~-2! !filesize:~4,2! !filesize:~2,2! !filesize:~0,2!
  echo 00 00 &rem Unused
  echo 00 00 &rem Unused
  echo 36 00 00 00 &rem Offset to the pixel array is 54 (14 bytes is the size of BITMAPFILEHEADER, 40 bytes is the size of BITMAPINFOHEADER)
  echo(
  rem BITMAPINFOHEADER https://docs.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-bitmapinfoheader
  echo 28 00 00 00 &rem 40 is the size of the BITMAPINFOHEADER
  echo !width:~-2! !width:~4,2! !width:~2,2! !width:~0,2!
  echo !height:~-2! !height:~4,2! !height:~2,2! !height:~0,2!
  echo 01 00 &rem One plane
  echo 18 00 &rem 24 bits per pixel
  echo 00 00 00 00 &rem 0 is for "uncompressed RGB bitmap"
  echo 00 00 00 00 &rem Size of the image which can be set to 0 for uncompressed bitmaps
  echo C4 0E 00 00 &rem Horizontal resolution in pixels per meter, set to 3780 (~96 dpi), might get ignored anyway
  echo C4 0E 00 00 &rem Vertical resolution in pixels per meter, set to 3780 (~96 dpi), might get ignored anyway
  echo 00 00 00 00 &rem Number of color indices in the color table (no color table present in a 24-bit bitmap because it's "true color" rather than "palette color")
  echo 00 00 00 00 &rem Number of color indices that are considered important for displaying the bitmap, 0 is for "every color is important", pointless here
  echo(
)
>nul copy /b "!filename!.hex.~tmp" + "!filename!.hex"
>nul move /y "!filename!.hex.~tmp" "!filename!.hex"
>nul certutil -decodehex -f "!filename!.hex" "!filename!.bmp" &rem Make the binary file out of the HEX data
del "!filename!.hex"
